import datetime as dt


__all__ = [
    "format_date_from_to",
]


def format_date_from_to(
    date_start: dt.date,
    date_end: dt.date,
    delimiter: str = "-",
) -> str:
    """
    Method to display date period.

    Possible output formats depending on the input data:

    * `DD.MM.YYYY - DD.MM.YYYY`
    * `DD.MM - DD.MM.YYYY`
    * `DD - DD.MM.YYYY`
    * `DD.MM.YYYY`

    Examples:
        ```pycon
        >>> print(format_date_from_to(dt.date(2020, 1, 1), dt.date(2020, 2, 1)))
        01.01 - 01.02.2020
        ```

    Attributes:
        date_start: Period start date.
        date_end: Period end date.
        delimiter: Dates delimiter.

    Returns:
        Date period string.
    """

    if date_start.year != date_end.year:
        return f"{date_start:%d.%m.%Y} {delimiter} {date_end:%d.%m.%Y}"

    if date_start.month != date_end.month:
        return f"{date_start:%d.%m} {delimiter} {date_end:%d.%m.%Y}"

    if date_start.day != date_end.day:
        return f"{date_start:%d} {delimiter} {date_end:%d.%m.%Y}"

    return f"{date_end:%d.%m.%Y}"



import warnings
from types import ModuleType
from typing import Any
from typing import Final

from django.conf import settings
from django.core.signals import setting_changed
from django.dispatch import receiver
from django.utils.module_loading import import_string

from meringue.conf import default_settings


SETTING_KEY: Final[str] = "MERINGUE"
"""
Parameter name in django settings for meringue settings.

Examples:
    ```py title="settings.py"
    MERINGUE = {
        "FRONTEND_URL": "http://meringue.local:9000/",
    }
    ```
"""


DEPRECATED_PARAMS: Final[dict[str, str]] = {}
"""
Dict with deprecated options and warning texts for them.

Examples:
    ```py
    DEPRECATED_PARAMS = {
        "PROTOCOL": "The `PROTOCOL` option is deprecated, use `BACKEND_PROTOCOL` instead.",
    }
    ```
"""

PARAMS_TO_IMPORT: Final[list[str]] = [
    "UPLOAD_RENAME_HANDLER",
]
"""
List of options that contain the path to the module and must be imported.

Examples:
    ```py
    PARAMS_TO_IMPORT = [
        "UPLOAD_RENAME_HANDLER",
    ]
    ```
"""


def import_from_string(val: str, attr: str) -> Any:
    """
    Imports a dotted module path and returns the attribute/class.

    Attributes:
        val: Dotted path to imported module.
        attr: The name of the parameter in the library settings.

    Raises:
        ImportError: Attribute/class not exists.

    Returns:
        Imported attribute/class.
    """

    try:
        return import_string(val)
    except ImportError as e:
        msg = f"Could not import '{val}' for API setting '{attr}'.\n{e.__class__.__name__}: {e}."
        raise ImportError(msg) from None


class Settings:
    """
    A settings object.

    The settings are obtained from the django settings by the name of the key which should contain
    an object with all the application settings.
    """

    def __init__(
        self,
        setting_key: str,
        defaults: dict[str, str],
        deprecated_params: dict[str, str] | None = None,
        params_to_impoprt: list[str] | None = None,
    ):
        """
        Attributes:
            setting_key: Settings key in django settings list.
            defaults: Dict with default parameter values. Used as a list of available settings.
            deprecated_params: Dict with deprecated options and warning texts for them.
            params_to_impoprt: List of options that contain the path to the module and must be
                imported.
        """

        self.setting_key = setting_key
        if isinstance(defaults, ModuleType):
            self.defaults = {}
            for key in dir(defaults):
                if key.isupper():
                    self.defaults[key] = getattr(defaults, key)
        else:
            self.defaults = defaults
        self.deprecated_params = deprecated_params or {}
        self.params_to_impoprt = params_to_impoprt or []
        self._cached_attrs = set()
        self.reset()

    def __getattr__(self, attr: str) -> Any:
        """
        Gets the parameter value and caches it in the attributes of the settings object.

        Attributes:
            attr: Setting parameter name.

        Raises:
            AttributeError: Error when trying to get an unregistered parameter.

        Warns:
            DeprecationWarning: A warning that the parameter is deprecated.

        Returns:
            Setting value.
        """
        if attr not in self.defaults:
            raise AttributeError("Invalid setting key: '%s'" % attr)

        if attr in self.deprecated_params:
            warnings.warn(self.deprecated_params[attr], DeprecationWarning, stacklevel=2)

        val = self.user_params.get(attr, self.defaults[attr])

        if attr in self.params_to_impoprt:
            val = import_from_string(val, attr)

        self._cached_attrs.add(attr)
        setattr(self, attr, val)
        return val

    def reset(self):
        """
        Reset downloaded settings, as well as clearing the cache.
        """

        for attr in self._cached_attrs:
            delattr(self, attr)
        self._cached_attrs = set()
        self.user_params = getattr(settings, self.setting_key, {})


m_settings = Settings(SETTING_KEY, default_settings, DEPRECATED_PARAMS, PARAMS_TO_IMPORT)


@receiver(setting_changed)
def reset_settings(*args, **kwargs):
    """
    Settings change signal handler.
    """

    if kwargs["setting"] == SETTING_KEY:
        m_settings.reset()



from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class Config(AppConfig):
    name = "meringue.core"
    verbose_name = _("Meringue")
    label = "meringue_core"



from rest_framework import exceptions
from rest_framework import views

from meringue.api.utils import render_error_details


def exception_handler(exc, context):
    """
    Error handler returning message and error code pairs.

    The handler is a wrapper over the standard handler `rest_framework.views.exception_handler`.
    """
    response = views.exception_handler(exc, context)

    if response is None:
        return response

    if isinstance(exc, exceptions.APIException):
        if isinstance(exc.detail, list):
            response.data = [render_error_details(e) for e in exc.detail]

        elif isinstance(exc.detail, dict):
            response.data = {f: render_error_details(e) for f, e in exc.detail.items()}

    return response



import base64

from Crypto.Cipher import AES

from meringue.conf import m_settings


__all__ = [
    "encrypt_message",
    "decrypt_message",
]


def encrypt_message(msg: str) -> str:
    """
    Method for encrypting a message with AES algorithm and the GCM method.

    Attributes:
        msg: Message to Encrypt.

    Returns:
        Encrypted message.
    """

    cipher = AES.new(m_settings.CRYPTO_KEY.encode("utf8"), AES.MODE_GCM)
    encrypted_msg = cipher.encrypt(msg.encode("utf8"))
    nonce = cipher.nonce
    result = base64.urlsafe_b64encode(nonce + encrypted_msg).decode("utf-8")
    return result


def decrypt_message(msg: str) -> str:
    """
    Method for decrypting a message encrypted with AES algorithm and the GCM method.

    Attributes:
        msg: Message to decrypt.

    Returns:
        Decrypted message.
    """
    raw_msg = base64.urlsafe_b64decode(msg.encode("utf-8"))
    nonce, raw_msg = raw_msg[: AES.block_size], raw_msg[AES.block_size :]
    cipher = AES.new(m_settings.CRYPTO_KEY.encode("utf-8"), AES.MODE_GCM, nonce=nonce)
    decrypted_msg = cipher.decrypt(raw_msg).decode("utf-8")
    return decrypted_msg



from rest_framework.exceptions import ErrorDetail


def render_error_details(error_detail: list | dict | ErrorDetail) -> list | dict:
    """
    The method renders errors of the ErrorDetail format into message and code pairs.

    Attributes:
        error_detail: Errors.

    Raises:
        Exception: Unknown error type.

    Returns:
        Rendered errors.
    """
    if isinstance(error_detail, list):
        error_list = [render_error_details(e) for e in error_detail]
        return error_list

    elif isinstance(error_detail, ErrorDetail):
        return {
            "message": str(error_detail),
            "code": error_detail.code,
        }

    elif isinstance(error_detail, dict):
        error_dict = {f: render_error_details(e) for f, e in error_detail.items()}
        return error_dict

    msg = f"Unknown error type {error_detail.__class__}: `{error_detail}`"
    raise Exception(msg)



from django.http import HttpResponse


def im_a_teapot(request):
    """
    We all at some point want to brew boiling water.

    Examples:
        ```py title="urls.py"
        from django.urls import path
        from meringue.core.views import im_a_teapot

        urlpatterns = [
            path('make_coffee', im_a_teapot, name="make_coffee"),
        ]
        ```
    """
    return HttpResponse(status=418)



from django.apps import apps

from modeltranslation.translator import TranslationOptions
from modeltranslation.translator import translator


for model in apps.get_models():
    fields = getattr(model._meta, "m_translate_fields", [])

    if fields and model not in translator.get_registered_models():
        translator.register(
            model,
            type(
                str(model.__name__ + "Translation"),
                (TranslationOptions,),
                {"fields": fields},
            ),
        )



from urllib.parse import urljoin

from meringue.conf import m_settings


__all__ = [
    "get_link",
]


def get_link(
    url_name: str,
    absolute: bool | None = True,
    **kwargs,
) -> str:
    """
    Method for getting a link.

    Attributes:
        url_name: The url name
        absolute: Flag to make URL absolute or relative
        kwargs: Url parameters

    Raises:
        Exception: FRONTEND_URLS parameter is empty
        Exception: The passed link is not specified in the FRONTEND_URLS parameter
        Exception: FRONTEND_DOMAIN parameter not specified

    Returns:
        link to the front.
    """

    if m_settings.FRONTEND_URLS is None:
        msg = "FRONTEND_URLS parameter is empty"
        raise Exception(msg)

    if url_name not in m_settings.FRONTEND_URLS:
        msg = f"The passed link `{url_name}` is not specified in the FRONTEND_URLS parameter"
        raise Exception(msg)

    url = m_settings.FRONTEND_URLS[url_name]
    url = url(**kwargs) if callable(url) else url.format(**kwargs)

    if absolute:
        if m_settings.FRONTEND_DOMAIN is None:
            msg = "FRONTEND_DOMAIN option not set"
            raise Exception(msg)

        url = urljoin(m_settings.FRONTEND_DOMAIN, url)

    return url



# ruff: noqa: RUF012

from django.db import models
from django.utils.translation import gettext_lazy as _

from meringue.core.query import UniversalQuerySet


class CMTimeMixin(models.Model):
    """
    A simple mixin to add _ctime_ and _mtime_ fields.
    """

    ctime = models.DateTimeField(auto_now_add=True, help_text=_("Date and time of creation."))
    mtime = models.DateTimeField(auto_now=True, help_text=_("Date and time of editing."))

    class Meta:
        abstract = True


class SortingMixin(models.Model):
    """
    Simple mixin to add sorting field.
    """

    sorting = models.SmallIntegerField(
        verbose_name=_("Sorting"),
        help_text=_("Sorting order."),
        default=0,
    )

    objects = UniversalQuerySet.as_manager()

    class Meta:
        ordering = [
            "sorting",
        ]
        abstract = True


class PublicationMixin(models.Model):
    """
    Mixin with the functionality of manual publishing.

    Examples:
        >>> FooModel.object.published()
        >>> FooModel.object.unpublished()
    """

    is_published = models.BooleanField(
        verbose_name=_("Publication"),
        help_text=_("Show/Hide"),
        default=True,
        db_index=True,
    )

    objects = UniversalQuerySet.as_manager()

    class Meta:
        abstract = True


class PublicationDatesMixin(models.Model):
    """
    Mixin with the functionality of publishing in a certain period.

    Examples:
        >>> FooModel.object.published()
        >>> FooModel.object.unpublished()
    """

    date_from = models.DateTimeField(
        verbose_name=_("Date from"),
        help_text=_("Date and time of publication (inclusive)."),
        blank=True,
        null=True,
    )
    date_to = models.DateTimeField(
        verbose_name=_("Date to"),
        help_text=_("Date and time when to hide."),
        blank=True,
        null=True,
    )

    objects = UniversalQuerySet.as_manager()

    class Meta:
        abstract = True



from typing import Final

from django.conf import settings


# CORE ############################################################################################

UPLOAD_RENAME_HANDLER: Final[str] = "meringue.core.upload_handlers.rename_handler"
"""
Path to method for renaming images on upload
"""

COP_YEAR: Final[int] = None
"""
Project start year for the copyright tag
"""

COP_YEARS_DIFF: Final[int] = 10
"""
Difference in years for which it is necessary to display the range of years
"""

CRYPTO_KEY: Final[str] = settings.SECRET_KEY[:32]
"""
Encryption key
"""

FRONTEND_URLS: Final[dict] = None
"""
A dict of links to the frontend
"""

FRONTEND_DOMAIN: Final[str] = None
"""
Domain for generating absolute links
"""


# API #############################################################################################

API_ENABLE_ROOT_VIEW: Final[str] = settings.DEBUG
"""
Option to enable or disable the root view of the [Router][meringue.api.routers.DefaultRouter]
"""
